import torch
import torch.nn as nn
from gaussian_splatting import GaussianModel


class VariableSHGaussianModel(GaussianModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._degrees = torch.empty(0)

    def to(self, device):
        self._degrees = self._degrees.to(device)
        return super().to(device)

    @property
    def get_features(self):
        features_dc = self._features_dc
        # compute SH according to self._degrees
        n_SH = (self._degrees + 1) ** 2 - 1
        indices = torch.arange((self.max_sh_degree + 1) ** 2 - 1, device=n_SH.device).expand(n_SH.shape[0], -1) < n_SH.unsqueeze(-1)
        features_rest = torch.zeros_like(self._features_rest)
        features_rest[indices, :] = self._features_rest[indices, :]
        with torch.no_grad():
            self._features_rest[~indices, :] = 0
            if self._features_rest.grad is not None:
                self._features_rest.grad[~indices, :] = 0
        return torch.cat((features_dc, features_rest), dim=1)

    def init_degrees(self):
        self._degrees = torch.zeros(self._xyz.shape[0], dtype=torch.int, device=self._xyz.device) + self.max_sh_degree

    def create_from_pcd(self, *args, **kwargs):
        super().create_from_pcd(*args, **kwargs)
        self.init_degrees()

    def load_ply(self, *args, **kwargs):
        super().load_ply(*args, **kwargs)
        self.init_degrees()

    def update_points_add(
            self,
            xyz: nn.Parameter,
            features_dc: nn.Parameter,
            features_rest: nn.Parameter,
            scaling: nn.Parameter,
            rotation: nn.Parameter,
            opacity: nn.Parameter,
    ):
        super().update_points_add(
            xyz=xyz,
            features_dc=features_dc,
            features_rest=features_rest,
            scaling=scaling,
            rotation=rotation,
            opacity=opacity,
        )
        with torch.no_grad():
            self._degrees = torch.cat((self._degrees, torch.zeros(xyz.shape[0]-self._degrees.shape[0], dtype=self._degrees.dtype, device=xyz.device) + self.max_sh_degree))

    def update_points_remove(
            self, removed_mask: torch.Tensor,
            xyz: nn.Parameter,
            features_dc: nn.Parameter,
            features_rest: nn.Parameter,
            scaling: nn.Parameter,
            rotation: nn.Parameter,
            opacity: nn.Parameter,):
        super().update_points_remove(
            removed_mask=removed_mask,
            xyz=xyz,
            features_dc=features_dc,
            features_rest=features_rest,
            scaling=scaling,
            rotation=rotation,
            opacity=opacity,
        )
        with torch.no_grad():
            self._degrees = self._degrees[~removed_mask]



import os
import shutil
from gaussian_splatting import GaussianModel
from reduced_3dgs.quantization import ExcludeZeroSHQuantizer as VectorQuantizer


def copy_not_exists(source, destination):
    if os.path.exists(destination):
        if os.path.samefile(source, destination):
            return
        os.remove(destination)
    shutil.copy(source, destination)


def quantize(source, destination, iteration, sh_degree, device, **kwargs):
    copy_not_exists(os.path.join(source, "cfg_args"), os.path.join(destination, "cfg_args"))
    copy_not_exists(os.path.join(source, "cameras.json"), os.path.join(destination, "cameras.json"))

    input = os.path.join(source, "point_cloud", "iteration_" + str(iteration), "point_cloud.ply")
    output = os.path.join(destination, "point_cloud", "iteration_" + str(iteration), "point_cloud_quantized.ply")
    os.makedirs(os.path.join(destination, "point_cloud", "iteration_" + str(iteration)), exist_ok=True)
    gaussians = GaussianModel(sh_degree).to(device)
    gaussians.load_ply(input)
    quantizer = VectorQuantizer(gaussians, **kwargs)
    quantizer.save_quantized(output)
    quantizer.quantize()
    quantizer.load_quantized(output)
    output = os.path.join(destination, "point_cloud", "iteration_" + str(iteration), "point_cloud.ply")
    gaussians.save_ply(output)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--source", required=True, type=str)
    parser.add_argument("-d", "--destination", required=True, type=str)
    parser.add_argument("-i", "--iteration", default=30000, type=int)
    parser.add_argument("--sh_degree", type=int, default=3)
    parser.add_argument("--device", type=str, default="cuda")
    parser.add_argument("--num_clusters", type=int, default=256)
    parser.add_argument("--num_clusters_rotation_re", type=int, default=None)
    parser.add_argument("--num_clusters_rotation_im", type=int, default=None)
    parser.add_argument("--num_clusters_opacity", type=int, default=None)
    parser.add_argument("--num_clusters_scaling", type=int, default=None)
    parser.add_argument("--num_clusters_features_dc", type=int, default=None)
    parser.add_argument("--num_clusters_features_rest", nargs="+", type=int, default=[])
    args = parser.parse_args()
    quantize(**vars(args))



from .abc import AbstractQuantizer, QuantizeTrainerWrapper
from .quantizer import VectorQuantizer
from .exclude_zeros import ExcludeZeroSHQuantizer
from .wrapper import VectorQuantizeTrainerWrapper, BaseVectorQuantizeTrainer



import torch
from gaussian_splatting import GaussianModel
from .quantizer import VectorQuantizer


class ExcludeZeroSHQuantizer(VectorQuantizer):
    def __init__(self, model: GaussianModel, *args, treat_as_zero=1e-8, **kwargs):
        super(ExcludeZeroSHQuantizer, self).__init__(model, *args, **kwargs)
        self.treat_as_zero = treat_as_zero

    def zeros_mask(self, values: torch.Tensor):
        return (values.abs() < self.treat_as_zero).all(-1)

    def generate_codebook_exclude_zero(self, values: torch.Tensor, num_clusters=256, init_codebook=None):
        zeros_mask = self.zeros_mask(values)
        if zeros_mask.all():
            return torch.zeros(1, values.shape[1], dtype=values.dtype, device=values.device), torch.zeros(values.shape[0], dtype=torch.long, device=values.device)
        if init_codebook is not None:
            if init_codebook.abs().max() < self.treat_as_zero:
                init_codebook = None
            elif init_codebook.shape[0] > num_clusters - 1:
                init_codebook = init_codebook[-(num_clusters - 1):, ...]
        nonzero_values = values[~zeros_mask]
        nonzero_centers, nonzero_ids = super().generate_codebook(nonzero_values, num_clusters - 1, init_codebook)
        ids = torch.zeros(values.shape[0], dtype=nonzero_ids.dtype, device=nonzero_ids.device)
        ids[~zeros_mask] = nonzero_ids + 1
        centers = torch.cat((torch.zeros(1, values.shape[1], dtype=values.dtype, device=values.device), nonzero_centers), dim=0)
        return centers, ids

    def has_zero(self, values: torch.Tensor):
        return self.zeros_mask(values).any()

    def produce_clusters_degree_features_rest(self, sh_degree, *args, **kwargs):
        features_rest_flatten = self.model._features_rest.detach().transpose(1, 2).flatten(0, 1)
        sh_idx_start, sh_idx_end = (sh_degree + 1) ** 2 - 1, (sh_degree + 2) ** 2 - 1
        features_rest = features_rest_flatten[:, sh_idx_start:sh_idx_end]
        if self.has_zero(features_rest):
            codebook, ids = self.generate_codebook_exclude_zero(features_rest, self.num_clusters_features_rest[sh_degree], *args, **kwargs)
        else:
            codebook, ids = self.generate_codebook(features_rest, self.num_clusters_features_rest[sh_degree], *args, **kwargs)
        return codebook, ids.reshape(-1, self.model._features_rest.shape[-1])



import abc

import torch

from gaussian_splatting import GaussianModel
from gaussian_splatting.trainer import AbstractTrainer, TrainerWrapper


class AbstractQuantizer(abc.ABC):

    @property
    @abc.abstractmethod
    def model(self) -> GaussianModel:
        raise ValueError("Model is not set")

    @abc.abstractmethod
    def quantize(self):
        pass

    @abc.abstractmethod
    def save_quantized(self, model: GaussianModel, ply_path: str, codebook_path: str):
        pass

    @abc.abstractmethod
    def load_quantized(self, model: GaussianModel, ply_path: str, codebook_path: str):
        pass


class QuantizeTrainerWrapper(TrainerWrapper, metaclass=abc.ABCMeta):
    def __init__(
            self, base_trainer: AbstractTrainer,
            quantizer: AbstractQuantizer,
            quantizate_from_iter=5000,
            quantizate_until_iter=30000,
            quantizate_interval=1000,
    ):
        super().__init__(base_trainer)
        self.quantizer = quantizer
        self.quantizate_from_iter = quantizate_from_iter
        self.quantizate_until_iter = quantizate_until_iter
        self.quantizate_interval = quantizate_interval

    @property
    def model(self) -> GaussianModel:
        model = self.base_trainer.model
        if self.quantizate_from_iter <= self.curr_step < self.quantizate_until_iter and self.curr_step % self.quantizate_interval == 0:
            with torch.no_grad():
                self.quantizer.quantize()
        return model



from typing import List
import torch

from gaussian_splatting import Camera
from gaussian_splatting.trainer import AbstractTrainer, TrainerWrapper, BaseTrainer
from gaussian_splatting.dataset import CameraDataset
from reduced_3dgs.diff_gaussian_rasterization._C import calculate_colours_variance

from .gaussian_model import VariableSHGaussianModel


def _low_variance_colour_culling(self: VariableSHGaussianModel, threshold, weighted_variance: torch.Tensor, weighted_mean: torch.Tensor):
    original_degrees = torch.zeros_like(self._degrees)
    original_degrees.copy_(self._degrees)

    # Uniform colour culling
    weighted_colour_std = weighted_variance.sqrt()
    weighted_colour_std[weighted_colour_std.isnan()] = 0
    weighted_colour_std = weighted_colour_std.mean(dim=2).squeeze()

    std_mask = weighted_colour_std < threshold
    self._features_dc[std_mask] = (weighted_mean[std_mask] - 0.5) / 0.28209479177387814
    self._degrees[std_mask] = 0
    self._features_rest[std_mask] = 0


def _low_distance_colour_culling(self: VariableSHGaussianModel, threshold, colour_distances: torch.Tensor):
    colour_distances[colour_distances.isnan()] = 0

    # Loop from active_sh_degree - 1 to 0, since the comparisons
    # are always done based on the max band that corresponds to active_sh_degree
    for sh_degree in range(self.active_sh_degree - 1, 0, -1):
        coeffs_num = (sh_degree+1)**2 - 1
        mask = colour_distances[:, sh_degree] < threshold
        self._degrees[mask] = torch.min(
            torch.tensor([sh_degree], device="cuda", dtype=int),
            self._degrees[mask]
        ).int()

        # Zero-out the associated SH coefficients for clarity,
        # as they won't be used in rasterisation due to the degrees field
        self._features_rest[mask, coeffs_num:] = 0


def _cull_sh_bands(self: VariableSHGaussianModel, cameras: List[Camera], threshold=0, std_threshold=0.):
    camera_positions = torch.stack([cam.camera_center for cam in cameras], dim=0)
    camera_viewmatrices = torch.stack([cam.world_view_transform for cam in cameras], dim=0)
    camera_projmatrices = torch.stack([cam.full_proj_transform for cam in cameras], dim=0)
    camera_fovx = torch.tensor([camera.FoVx for camera in cameras], device="cuda", dtype=torch.float32)
    camera_fovy = torch.tensor([camera.FoVy for camera in cameras], device="cuda", dtype=torch.float32)
    image_height = torch.tensor([camera.image_height for camera in cameras], device="cuda", dtype=torch.int32)
    image_width = torch.tensor([camera.image_width for camera in cameras], device="cuda", dtype=torch.int32)

    # Wrapping in a function since it's called with the same parameters twice
    def run_calculate_colours_variance():
        return calculate_colours_variance(
            camera_positions,
            self.get_xyz,
            self._opacity,
            self.get_scaling,
            self.get_rotation,
            camera_viewmatrices,
            camera_projmatrices,
            torch.tan(camera_fovx*0.5),
            torch.tan(camera_fovy*0.5),
            image_height,
            image_width,
            self.get_features,
            self._degrees,
            self.active_sh_degree)

    _, weighted_variance, weighted_mean = run_calculate_colours_variance()
    _low_variance_colour_culling(self, std_threshold, weighted_variance, weighted_mean)

    # Recalculate to account for the changed values
    colour_distances, _, _ = run_calculate_colours_variance()
    _low_distance_colour_culling(self, threshold, colour_distances)


def cull_sh_bands(self: VariableSHGaussianModel, cameras: List[Camera], threshold=0, std_threshold=0.):
    with torch.no_grad():
        _cull_sh_bands(self, cameras, threshold, std_threshold)


class SHCuller(TrainerWrapper):
    def __init__(
            self, base_trainer: AbstractTrainer,
            dataset: CameraDataset,
            cdist_threshold: float = 6,
            std_threshold: float = 0.04,
            cull_at_steps=[15000],
    ):
        super().__init__(base_trainer)
        assert isinstance(self.model, VariableSHGaussianModel)
        self.dataset = dataset
        self.cdist_threshold = cdist_threshold
        self.std_threshold = std_threshold
        self.cull_at_steps = cull_at_steps

    def optim_step(self):
        ret = super().optim_step()
        if self.curr_step in self.cull_at_steps:
            cull_sh_bands(self.model, self.dataset, self.cdist_threshold, self.std_threshold)
        return ret


def BaseSHCullingTrainer(
    model: VariableSHGaussianModel,
        scene_extent: float,
        dataset: CameraDataset,
        cdist_threshold: float = 6,
        std_threshold: float = 0.04,
        cull_at_steps=[15000],
        *args, **kwargs):
    return SHCuller(
        BaseTrainer(model, scene_extent, *args, **kwargs),
        dataset,
        cdist_threshold=cdist_threshold,
        std_threshold=std_threshold,
        cull_at_steps=cull_at_steps,
    )



from typing import List
import torch
from gaussian_splatting import GaussianModel, Camera
from gaussian_splatting.trainer import AbstractDensifier, Densifier, DensificationInstruct, DensificationTrainer
from reduced_3dgs.diff_gaussian_rasterization._C import sphere_ellipsoid_intersection, allocate_minimum_redundancy_value, find_minimum_projected_pixel_size
from reduced_3dgs.simple_knn._C import distIndex2


def calculate_redundancy_metric(gaussians: GaussianModel, cameras: List[Camera], pixel_scale=1.0, num_neighbours=30):
    # Get minimum projected pixel size
    cube_size = find_minimum_projected_pixel_size(
        torch.stack([camera.full_proj_transform for camera in cameras], dim=0),
        torch.stack([camera.full_proj_transform.inverse() for camera in cameras], dim=0),
        gaussians._xyz,
        torch.tensor([camera.image_height for camera in cameras], device="cuda", dtype=torch.int32),
        torch.tensor([camera.image_width for camera in cameras], device="cuda", dtype=torch.int32)
    )

    scaled_pixel_size = cube_size * pixel_scale
    half_diagonal = scaled_pixel_size * torch.sqrt(torch.tensor([3], device="cuda")) / 2

    # Find neighbours as candidates for the intersection test
    _, indices = distIndex2(gaussians.get_xyz, num_neighbours)
    indices = indices.view(-1, num_neighbours)

    # Do the intersection check
    redundancy_metrics, intersection_mask = sphere_ellipsoid_intersection(gaussians._xyz,
                                                                          gaussians.get_scaling,
                                                                          gaussians.get_rotation,
                                                                          indices,
                                                                          half_diagonal,
                                                                          num_neighbours)
    # We haven't counted count for the primitive at the center of each sphere, so add 1 to everything
    redundancy_metrics += 1

    indices = torch.cat((torch.arange(gaussians.get_xyz.shape[0], device="cuda", dtype=torch.int).view(-1, 1), indices), dim=1)
    intersection_mask = torch.cat((torch.ones_like(gaussians._opacity, device="cuda", dtype=bool), intersection_mask), dim=1)

    min_redundancy_metrics = allocate_minimum_redundancy_value(redundancy_metrics, indices, intersection_mask, num_neighbours+1)[0]
    return min_redundancy_metrics, cube_size


def mercy_points(self: GaussianModel, _splatted_num_accum: torch.Tensor, lambda_mercy=2, mercy_minimum=2, mercy_type='redundancy_opacity'):
    mean = _splatted_num_accum.float().mean(dim=0, keepdim=True)
    std = _splatted_num_accum.float().var(dim=0, keepdim=True).sqrt()

    threshold = max((mean + lambda_mercy*std).item(), mercy_minimum)

    mask = (_splatted_num_accum > threshold)

    if mercy_type == 'redundancy_opacity':
        # Prune redundant points based on 50% lowest opacity
        mask[mask.clone()] = self.get_opacity[mask].squeeze() < self.get_opacity[mask].median()
    elif mercy_type == 'redundancy_random':
        # Prune 50% redundant points at random
        mask[mask.clone()] = torch.rand(mask[mask].shape, device="cuda").squeeze() < 0.5
    elif mercy_type == 'opacity':
        # Prune based just on opacity
        threshold = self.get_opacity.quantile(0.045)
        mask = (self.get_opacity < threshold).squeeze()
    elif mercy_type == 'redundancy_opacity_opacity':
        # Prune based on opacity and on redundancy + opacity (options 1 and 3)
        mask[mask.clone()] = self.get_opacity[mask].squeeze() < self.get_opacity[mask].median()
        threshold = torch.min(self.get_opacity.quantile(0.03), torch.tensor([0.05], device="cuda"))
        mask = torch.logical_or(mask, (self.get_opacity < threshold).squeeze())
    return mask


def mercy_gaussians(
    model: GaussianModel,
    dataset: List[Camera],
    box_size=1.,
    lambda_mercy=1.,
    mercy_minimum=3,
    mercy_type='redundancy_opacity'
):
    _splatted_num_accum, _ = calculate_redundancy_metric(model, dataset, pixel_scale=box_size)
    mask = mercy_points(model, _splatted_num_accum.squeeze(), lambda_mercy, mercy_minimum, mercy_type)
    return mask


class BasePruner(AbstractDensifier):
    def __init__(
            self, model: GaussianModel, dataset: List[Camera],
            prune_from_iter=1000,
            prune_until_iter=15000,
            prune_interval: int = 100,
            box_size=1.,
            lambda_mercy=1.,
            mercy_minimum=3,
            mercy_type='redundancy_opacity'):
        self._model = model
        self.dataset = dataset
        self.prune_from_iter = prune_from_iter
        self.prune_until_iter = prune_until_iter
        self.prune_interval = prune_interval
        self.box_size = box_size
        self.lambda_mercy = lambda_mercy
        self.mercy_minimum = mercy_minimum
        self.mercy_type = mercy_type

    @property
    def model(self) -> GaussianModel:
        return self._model

    def densify_and_prune(self, loss, out, camera, step: int) -> DensificationInstruct:
        if self.prune_from_iter < step < self.prune_until_iter and step % self.prune_interval == 0:
            return DensificationInstruct(remove_mask=mercy_gaussians(self.model, self.dataset, self.box_size, self.lambda_mercy, self.mercy_minimum, self.mercy_type))
        return DensificationInstruct()


def BasePruningTrainer(
        model: GaussianModel,
        scene_extent: float,
        dataset: List[Camera],
        prune_from_iter=1000,
        prune_until_iter=15000,
        prune_interval: int = 100,
        box_size=1.,
        lambda_mercy=1.,
        mercy_minimum=3,
        mercy_type='redundancy_opacity',
        *args, **kwargs):
    return DensificationTrainer(
        model, scene_extent,
        BasePruner(
            model, dataset,
            prune_from_iter, prune_until_iter, prune_interval,
            box_size, lambda_mercy, mercy_minimum, mercy_type
        ), *args, **kwargs
    )


class PrunerInDensify(Densifier):
    def __init__(
            self, model: GaussianModel, scene_extent, dataset: List[Camera],
            box_size=1.,
            lambda_mercy=1.,
            mercy_minimum=3,
            mercy_type='redundancy_opacity',
            *args, **kwargs):
        super().__init__(model, scene_extent, *args, **kwargs)
        self.dataset = dataset
        self.box_size = box_size
        self.lambda_mercy = lambda_mercy
        self.mercy_minimum = mercy_minimum
        self.mercy_type = mercy_type

    def prune(self) -> torch.Tensor:
        _splatted_num_accum, _ = calculate_redundancy_metric(self.model, self.dataset, pixel_scale=self.box_size)
        mask = mercy_points(self.model, _splatted_num_accum.squeeze(), self.lambda_mercy, self.mercy_minimum, self.mercy_type)
        return mask


def BasePrunerInDensifyTrainer(
        model: GaussianModel,
        scene_extent: float,

        dataset: List[Camera],
        box_size=1.,
        lambda_mercy=1.,
        mercy_minimum=3,
        mercy_type='redundancy_opacity',

        percent_dense=0.01,

        densify_from_iter=500,
        densify_until_iter=15000,
        densify_interval=100,
        densify_grad_threshold=0.0002,
        densify_opacity_threshold=0.005,

        prune_from_iter=1000,
        prune_until_iter=15000,
        prune_interval=100,
        prune_screensize_threshold=20,

        *args, **kwargs):
    return DensificationTrainer(
        model, scene_extent,
        PrunerInDensify(
            model, scene_extent, dataset,
            box_size, lambda_mercy, mercy_minimum, mercy_type,
            percent_dense,
            densify_from_iter, densify_until_iter, densify_interval, densify_grad_threshold, densify_opacity_threshold,
            prune_from_iter, prune_until_iter, prune_interval, prune_screensize_threshold
        ), *args, **kwargs
    )
